def equ_to_dict(equation_str: str) -> dict:
    equation_str = equation_str.replace(' ','')
    equ_pair     = equation_str.split('=')
    if len(equ_pair) != 2:
        raise Exception("g_equations: Equation must have single-one '=' symbol")
    equation_dict        = dict()
    equation_dict["res"] = 0.
    equation_str         = equ_pair[0] + '='

    def go_coefficients(k: float) -> None:
        var_indx = 0
        for i in range(0,len(equation_str)):
            char = equation_str[i]
            if char != '+' and char != '-' and char != '=': continue
            if i == var_indx: continue
            var_str     = equation_str[var_indx:i]
            var_str_len = i - var_indx
            if var_str_len == 1 and var_indx != 0:
                raise Exception("g_equations: Loosing statement between math symbols")
            var_name = var_str[-1]
            if var_name.isalpha():
                var_value     = var_str[:-1]
                var_value_len = var_str_len - 1
                local_k       = k
            else:
                var_value     = var_str
                var_value_len = var_str_len
                var_name      = "res"
                local_k       = -k
            if var_value_len == 0 and var_indx == 0:
                var_value = 1.0
            elif var_value_len == 1 and var_indx != 0:
                if var_value[0] == '+': var_value =  1.0
                else:                   var_value = -1.0
            else:
                try:
                    var_value = float(var_value)
                except ValueError as e:
                    raise ValueError("g_equations: Variable {} has incorect coefficient - {}".format(var_name,e))
            if var_name in equation_dict:
                equation_dict[var_name] += var_value * local_k
            else:
                equation_dict[var_name] = var_value * local_k
            if char == '=': break 
            var_indx = i

    go_coefficients(1.)
    equation_str = equ_pair[1] + '='
    go_coefficients(-1.)
    return equation_dict

def dict_to_str(equation_dict: dict) -> str:
    out_str = ""
    for key in equation_dict.keys():
        if key == "res": continue
        var_name  = key
        var_value = equation_dict[key]
        if var_value == 0: continue
        out_str += " + " if var_value > 0 else " - "
        if abs(var_value) != 1.0: out_str += str(abs(var_value))
        out_str += var_name
    out_str += " = "
    try:
        out_str += str(equation_dict["res"])
    except KeyError:
        raise Exception("g_equations: dictionary must have value by 'res' key")
    return out_str[3:]


import numpy
import typing
import g_equations

class matrix_add_values_kwargs(typing.TypedDict,total=False):
    add_values: bool

def numpy_from_dicts(equations_dicts: list, **kwargs: matrix_add_values_kwargs) -> dict:
    matrix_triple = matrixs_from_dicts(equations_dicts,**kwargs)
    vars          = matrix_triple[2]
    A_matrix      = numpy.matrix(matrix_triple[0])
    B_matrix      = numpy.matrix(matrix_triple[1])
    A1_matrix     = numpy.linalg.inv(A_matrix)
    ans_matrix    = A1_matrix * B_matrix
    var_dict      = dict()
    for i in range(len(vars)):
        var_dict[vars[i]] = float(ans_matrix[i,0])
    return var_dict

def matrixs_from_dicts(equations_dicts: list, **kwargs: matrix_add_values_kwargs) -> tuple:
    add_values = kwargs.get("add_values",False)
    if add_values:
        keys = set()
        for dict in equations_dicts:
            for key in dict.keys():
                if key not in keys: keys.add(key)
        keys = sorted(keys)
        keys.remove("res")
    else:
        if not all(equations_dicts[i-1].keys() == equations_dicts[i].keys() for i in range(1,len(equations_dicts))):
            raise Exception("g_equations: Equations must have the same variables")
        keys = sorted(equations_dicts[0].keys())
        keys.remove("res")
    if len(keys) != len(equations_dicts):
        raise Exception("g_equations: Number of equations must be equal to number of variables")
    if add_values:
        A_matrix = [list(map(lambda key: e_d[key] if key in e_d.keys() else 0.0, keys)) for e_d in equations_dicts]
    else:
        A_matrix = [list(map(lambda key: e_d[key],keys)) for e_d in equations_dicts]
    B_matrix = [[e_d["res"]] for e_d in equations_dicts]
    return (A_matrix, B_matrix, tuple(keys))

def numpy_from_equs(equations_strs: list, **kwargs: matrix_add_values_kwargs) -> dict:
    return numpy_from_dicts(list(map(g_equations.equ_to_dict,equations_strs)),**kwargs)

roots_of_dicts = numpy_from_dicts
solve_equs     = numpy_from_equs

def matrixs_from_equs(equations_strs: list, **kwargs: matrix_add_values_kwargs) -> list:
    return matrixs_from_dicts(list(map(g_equations.equ_to_dict,equations_strs)),**kwargs)

def matrix_to_str(matrix: list) -> str:
    sttr = ""
    for line in matrix:
        line_str = ""
        for elem in line:
            line_str += ' ' + str(elem)
        sttr += line_str[1:] + "; "
    return sttr[:-2]